package build

import (
	"fmt"
	"strconv"

	"git.akyoto.dev/cli/q/src/build/asm"
	"git.akyoto.dev/cli/q/src/build/cpu"
	"git.akyoto.dev/cli/q/src/build/expression"
	"git.akyoto.dev/cli/q/src/build/fs"
	"git.akyoto.dev/cli/q/src/build/token"
	"git.akyoto.dev/cli/q/src/errors"
	"git.akyoto.dev/go/color/ansi"
)

// Function represents a function.
type Function struct {
	Name      string
	File      *fs.File
	Head      token.List
	Body      token.List
	Variables map[string]*Variable
	Assembler asm.Assembler
	CPU       cpu.CPU
	Error     error
	count     struct{ loop int }
}

// Compile turns a function into machine code.
func (f *Function) Compile() {
	f.Assembler.Label(f.Name)
	err := f.CompileTokens(f.Body)

	if err != nil {
		f.Error = err
		return
	}

	f.Assembler.Return()
}

// CompileTokens compiles a token list.
func (f *Function) CompileTokens(body token.List) error {
	start := 0
	groupLevel := 0
	blockLevel := 0

	for i, t := range body {
		if start == i && t.Kind == token.NewLine {
			start = i + 1
			continue
		}

		switch t.Kind {
		case token.NewLine:
			if groupLevel > 0 || blockLevel > 0 {
				continue
			}

			if start != -1 {
				instruction := body[start:i]
				err := f.CompileInstruction(instruction)

				if err != nil {
					return err
				}

				start = -1
			}

			start = i + 1

		case token.GroupStart:
			groupLevel++

		case token.GroupEnd:
			groupLevel--

		case token.BlockStart:
			blockLevel++

		case token.BlockEnd:
			blockLevel--
		}
	}

	return nil
}

// CompileInstruction compiles a single instruction.
func (f *Function) CompileInstruction(line token.List) error {
	if len(line) == 0 {
		return nil
	}

	if line[0].Kind == token.Keyword {
		return f.CompileKeyword(line)
	}

	expr := expression.Parse(line)

	if expr == nil {
		return nil
	}

	defer expr.Close()

	if isVariableDefinition(expr) {
		return f.CompileVariableDefinition(expr)
	}

	if isFunctionCall(expr) {
		return f.CompileFunctionCall(expr)
	}

	if expr.Token.Kind == token.Operator {
		switch expr.Token.Text() {
		case "+=":
			name := expr.Children[0].Token.Text()
			number, _ := strconv.Atoi(expr.Children[1].Token.Text())
			f.Assembler.AddRegisterNumber(f.Variables[name].Register, uint64(number))
			return nil
		}
	}

	return errors.New(&errors.InvalidInstruction{Instruction: expr.Token.Text()}, f.File, expr.Token.Position)
}

// ExpressionToRegister moves the result of an expression into the given register.
func (f *Function) ExpressionToRegister(root *expression.Expression, register cpu.Register) error {
	if root.IsLeaf() {
		return f.TokenToRegister(root.Token, register)
	}

	return errors.New(errors.NotImplemented, f.File, root.Token.Position)
}

// TokenToRegister moves a token into a register.
// It only works with identifiers, numbers and strings.
func (f *Function) TokenToRegister(t token.Token, register cpu.Register) error {
	switch t.Kind {
	case token.Identifier:
		name := t.Text()
		variable, exists := f.Variables[name]

		if !exists {
			return errors.New(&errors.UnknownIdentifier{Name: name}, f.File, t.Position)
		}

		f.Assembler.MoveRegisterRegister(register, variable.Register)
		return nil

	case token.Number:
		value := t.Text()
		n, err := strconv.Atoi(value)

		if err != nil {
			return err
		}

		f.Assembler.MoveRegisterNumber(register, uint64(n))
		return nil

	case token.String:
		return errors.New(errors.NotImplemented, f.File, t.Position)

	default:
		return errors.New(errors.InvalidExpression, f.File, t.Position)
	}
}

// PrintAsm shows the assembly instructions.
func (f *Function) PrintAsm() {
	ansi.Bold.Println(f.Name)
	ansi.Dim.Println("╭────────────────────────────────────────────────────────────")

	for _, x := range f.Assembler.Instructions {
		ansi.Dim.Print("│ ")
		fmt.Print(x.Mnemonic.String())

		if x.Data != nil {
			fmt.Print(" " + x.Data.String())
		}

		fmt.Print("\n")
	}

	ansi.Dim.Println("╰────────────────────────────────────────────────────────────")
}

// String returns the function name.
func (f *Function) String() string {
	return f.Name
}

// identifierExists returns true if the identifier has been defined.
func (f *Function) identifierExists(name string) bool {
	_, exists := f.Variables[name]
	return exists
}
